from enum import Enum


class Strictness_Level(Enum):
    """
        对于正确性与完整性的要求的严格程度
    """
    COMPLETE = "high"  # 所有节点均有一个或者多个匹配上的 backend，且第一个匹配上的 backend 就成功写入。
    COMPATIBLE = "normal"  # 所有节点均有一个或者多个匹配上的 backend，但是首先匹配到的 backend 写入出错，使用其后再次匹配到的其他 backend 能够成功写入
    # （这种情况更多应归咎于 backend 的 writable() 方法无法拒绝所有错误输入或者 backend 本身没有按照预期工作。一般而言这对最终写入内容的正确不会有太大影响。）
    # 这个等级是默认等级
    IGNORE_FAILURE = "low"  # 匹配不完整，或者某些节点尝试过所有匹配到的 backend 之后仍然无法写入



import numpy as np


def softmax(x, axis=-1, temperature=None, b_use_log_over_x=False):
    """
        一种将多个数值转换为概率分布的函数
            softmax(x)[k] = e^{x_k/t} / sum_{i}( e^{x_i/t} )

        参数：
            axis:               <int> 要对将哪一个轴转为概率分布（亦即沿着该轴求和将得到1）
                                    默认为最后一个轴
            temperature:        <float> 温度系数，起到对输入中的相对小/大值的抑制/增强作用。
                                    它是一个非负数。
                                    它越大，输出的概率分布越平缓。当趋于无穷时，输出退化为均匀分布。
                                    它越小，输出的概率分布越尖锐。当趋于0时，输出退化成狄拉克函数。
            b_use_log_over_x:   <boolean> 对输入的概率分布首先进行一次 log() 操作。
                                    默认为 False
                                    当设为 True 开启后，该函数将从 softmax 变为：
                                    softmax(log(x))[k] = x_k^{1/t} / sum_{i}( x_i^{1/t} )
    """
    x = np.asarray(x, dtype=float)

    if temperature == 0:
        # quick
        res = np.zeros_like(x)
        res = np.where(x == np.max(x, axis=axis), 1, res)
    elif b_use_log_over_x:
        # softmax(log(x))
        res = x ** (1 / temperature)
    else:
        # softmax(x)
        # 为了数值稳定，减去最大值
        x = x - np.max(x)
        #
        if temperature is not None:
            assert temperature > 0
            x /= temperature
        res = np.exp(x)
    res = res / np.sum(res, axis=axis, keepdims=True)
    return res


if __name__ == '__main__':
    print(softmax(np.asarray([0, 0.1]) * 10))
    print(softmax(np.asarray([0, 0.1]), temperature=0.1))
    print(softmax(np.asarray([[[0], [0.1]]]), temperature=0.00001, axis=1))
    print(softmax(np.asarray([[[0], [0.1]]]), temperature=0, axis=1))
    print(softmax([0, 1, 2], temperature=0.1))



import optuna


def build_sampler(name, seed=None, feasible_domain: dict = None, **kwargs):
    """
        构建超参数采样器实例

        参数：
            name：               <str> 算法的种类
                                    目前支持：
                                        - optuna.samplers 下的算法。
                                            比如要使用 TPE 算法，可以输入 "optuna.samplers.TPESampler"，或者直接简写成 "TPESampler"
            seed:               <int> 随机种子
                                    建议：对于多进程优化，不同进程所用的随机种子也应该不同，以产生不同的参数配置
            feasible_domain:    <dict> 对于 GridSampler 算法，可以通过指定该参数来生成算法所需的 search_space 参数。
            **kwargs:           其余参数将用于初始化算法实例
    """
    # 参数校验
    if not name.startswith("optuna"):
        name = "optuna.samplers." + name
    #
    if seed is not None:
        kwargs["seed"] = seed
    #
    if name.endswith("GridSampler"):
        # 需要补充 search_space 参数
        assert isinstance(feasible_domain, (dict,)), \
            f'for sampler {name}, feasible_domain(<dict>) must be specified to generate search_space'
        search_space = dict()
        for k, v in feasible_domain.items():
            assert v["p_type"] in ["categorical"], \
                f'for sampler {name}, feasible_domain(<dict>) should be categorical, but {k} is {v["p_type"]}'
            search_space[k] = v["choices"]
        #
        kwargs["search_space"] = search_space
    cls = eval(name)
    return cls(**kwargs)



import concurrent.futures
from kevin_toolbox.computer_science.data_structure import Executor


def multi_thread_execute(executors, thread_nums=50, b_display_progress=True, timeout=None, _hook_for_debug=None):
    """
        多线程执行

        参数：
            executors:                  <list/generator/iterator of Executor> 执行器序列
            thread_nums:                <int> 线程数
            b_display_progress:         <boolean> 是否显示进度条
            timeout:                    <int> 每个线程的最大等待时间，单位是s
                                            默认为 None，表示允许等待无限长的时间
            _hook_for_debug:            <dict/None> 当设置为非 None 值时，将保存中间的执行信息。
                                            包括：
                                                - "execution_order":    执行顺序
                                                - "completion_order":   完成顺序
                                            这些信息与最终结果无关，仅面向更底层的调试需求，任何人都不应依赖该特性
        返回：
            res_ls, failed_idx_ls
            执行结果列表，执行失败的执行器 idx
    """
    executor_ls = []
    for i in executors:
        assert isinstance(i, (Executor,))
        executor_ls.append(i)
    if b_display_progress:
        from tqdm import tqdm
        p_bar = tqdm(total=len(executor_ls))
    else:
        p_bar = None
    _execution_orders, _completion_orders = [], []

    def wrapper(executor, idx):
        nonlocal p_bar, _execution_orders, _completion_orders
        _execution_orders.append(idx)
        res = executor.run()
        _completion_orders.append(idx)
        if p_bar is not None:
            p_bar.update()

        return res

    res_ls, failed_idx_ls = [], []
    with concurrent.futures.ThreadPoolExecutor(max_workers=thread_nums) as thread_pool:
        # 提交任务
        futures = [thread_pool.submit(wrapper, executor, i) for i, executor in enumerate(executors)]
        # 设置超时时间
        concurrent.futures.wait(futures, timeout=timeout)
        # 
        for i, future in enumerate(futures):
            if future.done() and not future.cancelled():
                res_ls.append(future.result())
            else:
                res_ls.append(None)
                failed_idx_ls.append(i)
    if b_display_progress:
        p_bar.close()

    #
    if isinstance(_hook_for_debug, (dict,)):
        _hook_for_debug.update(dict(execution_orders=_execution_orders, completion_orders=_completion_orders))

    return res_ls, failed_idx_ls


if __name__ == '__main__':
    import time


    def func_(i):
        if i in [2, 3, 7]:
            time.sleep(10)
        else:
            time.sleep(2)
        print(i)
        return i * 2


    hook_for_debug = dict()
    print(multi_thread_execute(executors=[Executor(func=func_, args=(i,)) for i in range(10)], thread_nums=5,
                               _hook_for_debug=hook_for_debug))
    print(hook_for_debug)



import optuna
import kevin_toolbox.nested_dict_list as ndl
from kevin_toolbox.patches.for_optuna.serialize import for_trial
from kevin_toolbox.env_info import version


def __converter(_, v):
    if isinstance(v, optuna.storages.BaseStorage):
        return dict(name=f'optuna.storages.{v.__class__.__name__}', url=getattr(v, "url", None))
    elif isinstance(v, optuna.samplers.BaseSampler):
        return dict(name=f'optuna.samplers.{v.__class__.__name__}')
    elif isinstance(v, optuna.pruners.BasePruner):
        return dict(name=f'optuna.pruners.{v.__class__.__name__}')
    elif isinstance(v, optuna.study.StudyDirection):
        return dict(name=f'optuna.study.StudyDirection', value=v.value, description=v.name)
    elif isinstance(v, optuna.trial.BaseTrial):
        return for_trial.dump(trial=v)
    return v


def dump(study: optuna.study.Study):
    # 主要属性
    keys = ['best_params', 'best_trial', 'best_trials', 'best_value', 'direction', 'directions', 'trials', 'user_attrs']
    if version.compare(optuna.__version__, "<", "3.1.0"):
        keys.append('system_attrs')
    res_s = {k: getattr(study, k, None) for k in keys}

    # 其他信息
    res_s["__dict__"] = dict()
    for k, v in study.__dict__.items():
        if k.startswith("_"):
            k = k[1:]
        if k in ["thread_local", "optimize_lock"]:
            continue
        res_s["__dict__"][k] = v

    # 序列化
    res_s = ndl.traverse(var=res_s, match_cond=lambda _, __, v: not isinstance(v, (list, dict,)),
                         action_mode="replace", converter=__converter, b_use_name_as_idx=False,
                         b_traverse_matched_element=False)

    return res_s



from enum import Enum
from kevin_toolbox.nested_dict_list.name_handler import escape_node


class Action_Mode(Enum):
    REMOVE = "remove"
    REPLACE = "replace"
    SKIP = "skip"


class Traversal_Mode(Enum):
    DFS_PRE_ORDER = "dfs_pre_order"
    DFS_POST_ORDER = "dfs_post_order"
    BFS = "bfs"


def traverse(var, match_cond, action_mode="remove", converter=None,
             b_use_name_as_idx=False, traversal_mode="dfs_pre_order", b_traverse_matched_element=False):
    """
        遍历 var 找到符合 match_cond 的元素，将其按照 action_mode 指定的操作进行处理

        参数：
            var:                待处理数据
                                    当 var 不是 dict 或者 list 时，将直接返回 var 而不做处理
            match_cond:         <func> 元素的匹配条件
                                    函数类型为 def(parent_type, idx, value): ...
                                    其中：
                                        parent_type     该元素源自哪种结构体，有两个可能传入的值： list，dict
                                        idx             该元素在结构体中的位置
                                                            当 b_use_name_as_idx=False 时，
                                                                对于列表是 index，对于字典是 key
                                                            当为 True 时，传入的是元素在整体结构中的 name 位置，name的格式和含义参考
                                                                name_handler.parse_name() 中的介绍
                                        value           元素的值
            action_mode:        <str> 如何对匹配上的元素进行处理
                                    目前支持：
                                        "remove"        将该元素移除
                                        "replace"       将该元素替换为 converter() 处理后的结果
                                        "skip":         不进行任何操作
            converter:          <func> 参见 action_mode 中的 "replace" 模式
                                    函数类型为 def(idx, value): ...
                                    其中 idx 和 value 的含义参见参数 match_cond 介绍
            traversal_mode:     <str> 遍历的模式、顺序
                                    目前支持：
                                        "dfs_pre_order"         深度优先、先序遍历
                                        "dfs_post_order"        深度优先、后序遍历
                                        "bfs"                   宽度优先
                                    默认为 "dfs_pre_order"
            b_use_name_as_idx:  <boolean> 对于 match_cond/converter 中的 idx 参数，是传入整体的 name 还是父节点的 index 或 key。
                                    默认为 False
            b_traverse_matched_element  <boolean> 对于匹配上的元素，经过处理后，是否继续遍历该元素的内容
                                    默认为 False
    """
    assert callable(match_cond)
    action_mode = Action_Mode(action_mode)
    if action_mode is Action_Mode.REPLACE:
        assert callable(converter)
    traversal_mode = Traversal_Mode(traversal_mode)

    if traversal_mode is Traversal_Mode.BFS:
        return _bfs(var, match_cond, action_mode, converter, b_use_name_as_idx, b_traverse_matched_element)
    else:
        return _dfs(var, match_cond, action_mode, converter, b_use_name_as_idx, traversal_mode,
                    b_traverse_matched_element, "")


def _bfs(var, match_cond, action_mode, converter, b_use_name_as_idx, b_traverse_matched_element):
    temp = [("", var)]

    while len(temp):
        pre_name, i = temp.pop(0)
        if isinstance(i, (list, dict)):
            keys = list(range(len(i)) if isinstance(i, list) else i.keys())
            keys.reverse()  # 反过来便于 列表 弹出元素
            idx_ls = _gen_idx(i, keys, b_use_name_as_idx, pre_name)

            # 匹配&处理
            for k, idx in zip(keys, idx_ls):
                b_matched, b_popped = _deal(i, k, idx, match_cond, converter, action_mode)
                if b_popped or (b_matched and not b_traverse_matched_element):
                    continue
                # 添加到队尾
                temp.append((idx, i[k]))

    return var


def _dfs(var, match_cond, action_mode, converter,
         b_use_name_as_idx, traversal_mode, b_traverse_matched_element, pre_name):
    if isinstance(var, (list, dict)):
        keys = list(range(len(var)) if isinstance(var, list) else var.keys())
        keys.reverse()  # 反过来便于 列表 弹出元素
        idx_ls = _gen_idx(var, keys, b_use_name_as_idx, pre_name)

        #
        if traversal_mode is Traversal_Mode.DFS_PRE_ORDER:
            # 先序
            # 匹配&处理
            deal_res_ls = []
            for k, idx in zip(keys, idx_ls):
                deal_res_ls.append(_deal(var, k, idx, match_cond, converter, action_mode))
            # 递归遍历
            for (b_matched, b_popped), k, idx in zip(deal_res_ls, keys, idx_ls):
                if b_popped or (b_matched and not b_traverse_matched_element):
                    continue
                var[k] = _dfs(var[k], match_cond, action_mode, converter, b_use_name_as_idx, traversal_mode,
                              b_traverse_matched_element, idx)
        else:
            # 后序
            # 递归遍历
            for k, idx in zip(keys, idx_ls):
                var[k] = _dfs(var[k], match_cond, action_mode, converter, b_use_name_as_idx, traversal_mode,
                              b_traverse_matched_element, idx)
            # 匹配&处理
            for k, idx in zip(keys, idx_ls):
                _deal(var, k, idx, match_cond, converter, action_mode)
    else:
        pass
    return var


def _deal(var, k, idx, match_cond, converter, action_mode):
    """处理节点"""
    # 匹配
    b_matched = match_cond(type(var), idx, var[k])
    b_popped = False
    # 处理
    if b_matched:
        if action_mode is Action_Mode.REMOVE:
            var.pop(k)
            b_popped = True
        elif action_mode is Action_Mode.REPLACE:
            var[k] = converter(idx, var[k])
        else:
            pass
    return b_matched, b_popped


def _gen_idx(var, keys, b_use_name_as_idx, pre_name):
    if b_use_name_as_idx:
        idx_ls = []
        for k in keys:
            method = "@" if isinstance(var, list) or not isinstance(k, str) else ":"
            k = escape_node(node=k, b_reversed=False, times=1)
            idx_ls.append(f'{pre_name}{method}{k}')
    else:
        idx_ls = keys
    return idx_ls



import os
from abc import ABC, abstractmethod


class Backend_Base(ABC):
    """
        对节点值进行序列化的抽象类

        使用方法：
            若要添加新的<序列化方式>，可以实现一个带有：
                - write(name, var, **kwargs):         序列化
                - read(name, **kwargs):               反序列化
                - writable(var, **kwargs):            是否可以写
                - readable(name, **kwargs):           是否可以读
            等方法的序列化 backend 类，然后将其实例注册到 SERIALIZER_BACKEND 中
    """

    def __init__(self, *args, **kwargs):
        """
            参数：
                folder：         <path> 保存、读取的文件夹
        """

        # 默认参数
        paras = {
            "folder": None,
        }

        # 获取参数
        paras.update(kwargs)

        # 校验参数
        assert isinstance(paras["folder"], (str,))
        os.makedirs(paras["folder"], exist_ok=True)

        self.paras = paras

    # ------------------------------------ 读取数据 ------------------------------------ #

    @abstractmethod
    def write(self, name, var, **kwargs):
        """
            序列化
        """
        assert self.writable(var=var)
        return

    @abstractmethod
    def read(self, name, **kwargs):
        """
            反序列化
        """
        assert self.readable(name=name)
        return

    @abstractmethod
    def writable(self, var, **kwargs):
        """
            是否可以写
        """
        return False

    @abstractmethod
    def readable(self, name, **kwargs):
        """
            是否可以写
        """
        return False



import torch
import numpy as np


def init_by_like(var):
    """
        构建一个与输入 var 具有相同类型、形状、设备的 0 数组

        参数：
            var:                <torch.tensor / np.ndarray / int / float>
    """
    if torch.is_tensor(var):
        res = torch.zeros_like(var)
    elif isinstance(var, (np.ndarray,)):
        res = np.zeros_like(var)
    elif isinstance(var, (int, float, np.number,)):
        res = 0.0
    else:
        raise ValueError("paras 'like' should be np.ndarray, torch.tensor or int/float")
    return res



def unescape_tuple(x):
    """
        将 tuple 进行反转义
            转义：     x ==> f"<eval>{x}"
            反转义：   f"<eval>{x}" ==> x

        为什么要进行转义？
            由于 json 中会将 tuple 作为 list 进行保存，因此在保存过程中会丢失相应信息。
    """
    if isinstance(x, str) and x.startswith("<eval>"):
        x = x[6:]
        if not x.startswith("<eval>"):
            x = eval(x)
        return x
    else:
        return x


if __name__ == '__main__':
    print(unescape_tuple("<eval>(1, 2, \"'1'\")"))
    # (1, 2, "\'1\'")
    print(unescape_tuple("<eval><eval>233"))
    # "<eval>233"



import os
from kevin_toolbox.computer_science.algorithm.registration import Registry

SERIALIZER_BACKEND = Registry(uid="SERIALIZER_BACKEND")

# 从 kevin_toolbox/nested_dict_list/serializer/backends 下收集被注册的 backend
SERIALIZER_BACKEND.collect_from_paths(path_ls=[os.path.join(os.path.dirname(__file__), "backends"), ],
                                      b_execute_now=False)



import asyncio
from tqdm import tqdm


def async_executor(func, paras_generator, b_use_tqdm=True):
    loop = asyncio.get_event_loop()
    res = loop.run_until_complete(_execute(func, paras_generator, b_use_tqdm))
    return res


async def _execute(func, paras_generator, b_use_tqdm):
    async def wrapper(*args, **kwargs):
        res = await func(*args, **kwargs)
        return res

    tasks = []
    if b_use_tqdm:
        paras_generator = tqdm(paras_generator)
    for paras in paras_generator:
        tasks.append(asyncio.ensure_future(wrapper(*paras.get("args", tuple()), **paras.get("kwargs", dict()))))

    res_ls = await asyncio.gather(*tasks)
    return res_ls


if __name__ == '__main__':
    async def func_(i):
        print(i)
        await asyncio.sleep(2)

        return i * 2


    print(async_executor(func=func_, paras_generator=[{"args": [i]} for i in range(1, 10)]))



import copy
from kevin_toolbox.nested_dict_list import traverse
import torch


def copy_(var, b_deepcopy=False, b_keep_internal_references=True):
    """
        复制嵌套字典列表 var，并返回其副本

        参数：
            var
            b_deepcopy:                     <boolean> 是否进行深拷贝
                                                默认为 False 此时仅复制结构，但叶节点仍在 var 和其副本之间共享
                                                当设置为 True 时，进行完全的深拷贝
            b_keep_internal_references:     <boolean> 是否保留内部的引用关系。
                                                默认为 True（当b_deepcopy=True时与 copy.deepcopy 的行为一致）。
                                什么是引用关系？
                                    比如我们将某个字典 A 多次加入到某个 list 中，那么这个 list 中的这些字典实际上都指向内存上同一个字典，
                                    因此对其中某个字典的修改将影响到其他 list 中的其他元素。这种内存上指向同一个位置的关系就是引用。
                                当使用 b_keep_internal_references=True 时，将保留 ndl 中结构与结构之间或者节点与节点之间的引用关系。
                                    当 b_deepcopy=False 进行浅拷贝时，该参数仅作用于结构，
                                    反之则同时作用于结构和节点。
        例子：
            以下面的结构为例（<xx>表示该结构体/节点内存中的地址）：
                {<0>
                    "a": [<1>
                        {<2> 1, 2, 3},
                        {<3> 4, 5, 6},
                        {<2> 1, 2, 3}
                    ],
                    "b": [<1>
                        {<2> 1, 2, 3},
                        {<3> 4, 5, 6},
                        {<2> 1, 2, 3}
                    ],
                }

            对于 b_deepcopy=False，
            当 b_keep_internal_references=True 时，生成：
                {<5>
                    "a": [<6>
                        {<2> 1, 2, 3},
                        {<3> 4, 5, 6},
                        {<2> 1, 2, 3}
                    ],
                    "b": [<6>
                        {<2> 1, 2, 3},
                        {<3> 4, 5, 6},
                        {<2> 1, 2, 3}
                    ],
                }
            当 b_keep_internal_references=False 时，生成：
                {<5>
                    "a": [<6>
                        {<2> 1, 2, 3},
                        {<3> 4, 5, 6},
                        {<2> 1, 2, 3}
                    ],
                    "b": [<7>  # !!!
                        {<2> 1, 2, 3},
                        {<3> 4, 5, 6},
                        {<2> 1, 2, 3}
                    ],
                }

            对于 b_deepcopy=True，
            当 b_keep_internal_references=True 时，生成：
                {<5>
                    "a": [<6>
                        {<8> 1, 2, 3},
                        {<9> 4, 5, 6},
                        {<8> 1, 2, 3}
                    ],
                    "b": [<6>
                        {<8> 1, 2, 3},
                        {<9> 4, 5, 6},
                        {<8> 1, 2, 3}
                    ],
                }
            当 b_keep_internal_references=False 时，生成：
                {<5>
                    "a": [<6>
                        {<8> 1, 2, 3},
                        {<9> 4, 5, 6},
                        {<8> 1, 2, 3}
                    ],
                    "b": [<7>  # !!!
                        {<10> 1, 2, 3},  # !!!
                        {<11> 4, 5, 6},  # !!!
                        {<10> 1, 2, 3}  # !!!
                    ],
                }
    """
    if b_deepcopy:
        if b_keep_internal_references:
            try:
                res = copy.deepcopy(var)  # 更快
                return res
            except:
                pass
        res = _copy_structure(var=var, b_keep_internal_references=b_keep_internal_references)
        res = _copy_nodes(var=res, b_keep_internal_references=b_keep_internal_references)
    else:
        res = _copy_structure(var=var, b_keep_internal_references=b_keep_internal_references)

    return res


def _copy_structure(var, b_keep_internal_references):
    """
        复制结构
            只复制 ndl 中的 dict、list 结构，复制前后 ndl 中的叶节点（亦即dict、list中的元素仍然保持共享）
    """
    memo_s = dict()  # {<old_id>: <new_it>, ...}

    def converter(_, value):
        if b_keep_internal_references:
            if id(value) not in memo_s:
                memo_s[id(value)] = value.copy()
            return memo_s[id(value)]
        else:
            return value.copy()

    return traverse(var=[var], match_cond=lambda _, __, value: isinstance(value, (list, dict,)),
                    action_mode="replace", converter=converter,
                    traversal_mode="dfs_pre_order", b_traverse_matched_element=True)[0]


def _copy_nodes(var, b_keep_internal_references):
    """
        复制叶节点
            复制并替换 ndl 中的叶节点
    """
    memo_s = dict()  # {<old_id>: <new_it>, ...}

    def copy_item(value):
        if torch.is_tensor(value):
            return value.detach().clone()
        else:
            return copy.deepcopy(value)

    def converter(_, value):
        if b_keep_internal_references:
            if id(value) not in memo_s:
                memo_s[id(value)] = copy_item(value)
            return memo_s[id(value)]
        else:
            return copy_item(value)

    return traverse(var=[var], match_cond=lambda _, __, value: not isinstance(value, (list, dict,)),
                    action_mode="replace", converter=converter,
                    traversal_mode="dfs_pre_order", b_traverse_matched_element=True)[0]


if __name__ == '__main__':
    x = dict(acc=[0.66, 0.78, 0.99], )
    copy_(var=x, b_deepcopy=False)



from kevin_toolbox.math.utils import spilt_integer_most_evenly


def generate_table(content_s, orientation="vertical", chunk_nums=None, chunk_size=None, b_allow_misaligned_values=False,
                   f_gen_order_of_values=None):
    """
        生成表格

        参数：
            content_s:              <dict> 内容
                                        支持两种输入模式：
                                            1.简易模式：
                                                content_s = {<title>: <list of value>, ...}
                                                此时键作为标题，值作为标题下的一系列值。
                                                由于字典的无序性，此时标题的顺序是不能保证的，若要额外指定顺序，请使用下面的 完整模式。
                                            2. 完整模式:
                                                content_s = {<index>: {"title": <title>,"values":<list of value>}, ...}
                                                此时将取第 <index> 个 "title" 的值来作为第 <index> 个标题的值。values 同理。
                                                该模式允许缺省某些 <index>，此时这些 <index> 对应的行/列将全部置空。
            orientation:            <str> 表格的方向
                                        支持以下值：
                                            "vertical" / "v":       纵向排列，亦即标题在第一行
                                            "horizontal" / "h":     横向排列，亦即标题在第一列
            chunk_nums:             <int> 将表格平均分割为多少份进行并列显示。
            chunk_size:             <int> 将表格按照最大长度进行分割，然后并列显示。
                注意：以上两个参数只能设置一个，同时设置时将报错
            b_allow_misaligned_values   <boolean> 允许不对齐的 values
                                        默认为 False，此时当不同标题下的 values 的长度不相等时，将会直接报错。
                                        当设置为 True 时，对于短于最大长度的 values 将直接补充 ""。
            f_gen_order_of_values:  <callable> 生成values排序顺序的函数
                                        该函数需要接受一个形如 {<title>: <value>, ...} 的 <dict>，并返回一个用于排序的 int/float/tuple
    """
    # 检验参数
    assert chunk_nums is None or 1 <= chunk_nums
    assert chunk_size is None or 1 <= chunk_size
    assert orientation in ["vertical", "horizontal", "h", "v"]

    # 将简易模式转换为完整模式
    if len(content_s.values()) > 0 and not isinstance(list(content_s.values())[0], (dict,)):
        content_s = {i: {"title": k, "values": v} for i, (k, v) in enumerate(content_s.items())}
    # 对齐 values
    len_ls = [len(v["values"]) for v in content_s.values()]
    max_length = max(len_ls)
    if min(len_ls) != max_length:
        assert b_allow_misaligned_values, \
            f'The lengths of the values under each title are not aligned. ' \
            f'The maximum length is {max_length}, but each length is {len_ls}'
        for v in content_s.values():
            v["values"].extend([""] * (max_length - len(v["values"])))
    # 对值进行排序
    if callable(f_gen_order_of_values):
        idx_ls = list(range(max_length))
        idx_ls.sort(key=lambda x: f_gen_order_of_values({v["title"]: v["values"][x] for v in content_s.values()}))
        for v in content_s.values():
            v["values"] = [v["values"][i] for i in idx_ls]
    # 补充缺省的 title
    for i in range(max(content_s.keys()) + 1):
        if i not in content_s:
            content_s[i] = {"title": "", "values": [""] * max_length}
    # 按照 chunk_nums 或者 chunk_size 对表格进行分割
    if chunk_nums is not None or chunk_size is not None:
        if chunk_nums is not None:
            split_len_ls = spilt_integer_most_evenly(x=max_length, group_nums=chunk_nums)
        else:
            split_len_ls = [chunk_size] * (max_length // chunk_size)
            if max_length % chunk_size != 0:
                split_len_ls += [max_length % chunk_size]
        max_length = max(split_len_ls)
        temp = dict()
        beg = 0
        for i, new_length in enumerate(split_len_ls):
            end = beg + new_length
            temp.update({k + i * len(content_s): {"title": v["title"],
                                                  "values": v["values"][beg:end] + [""] * (max_length - new_length)} for
                         k, v in content_s.items()})
            beg = end
        content_s = temp
    # 构建表格
    return _show_table(content_s=content_s, orientation=orientation)


def _show_table(content_s, orientation="vertical"):
    """
        生成表格

        参数：
            content_s:              <dict> 内容
                                        content_s = {<index>: {"title": <title>,"values":<list of value>}, ...}
                                        此时将取第 <index> 个 "title" 的值来作为第 <index> 个标题的值。values 同理。
            orientation:            <str> 表格的方向
                                        支持以下值：
                                            "vertical" / "v":       纵向排列，亦即标题在第一行
                                            "horizontal" / "h":     横向排列，亦即标题在第一列
    """
    table = ""
    if orientation in ["vertical", "v"]:
        table += "| " + " | ".join([f'{content_s[i]["title"]}' for i in range(len(content_s))]) + " |\n"
        table += "| " + " | ".join(["---"] * len(content_s)) + " |\n"
        for row in zip(*[content_s[i]["values"] for i in range(len(content_s))]):
            table += "| " + " | ".join([f'{i}' for i in row]) + " |\n"
    else:
        for i in range(len(content_s)):
            row = [f'{content_s[i]["title"]}'] + [f'{i}' for i in content_s[i]["values"]]
            table += "| " + " | ".join(row) + " |\n"
            if i == 0:
                table += "| " + " | ".join(["---"] * len(row)) + " |\n"
    return table


if __name__ == '__main__':
    content_s = {0: dict(title="a", values=[1, 2, 3]), 2: dict(title="b", values=[4, 5, 6])}
    doc = generate_table(content_s=content_s, orientation="h")
    print(doc)

    # from collections import OrderedDict
    #
    # content_s = OrderedDict({
    #     "y/n": [True] * 5 + [False] * 5,
    #     "a": list(range(10)),
    #     "b": [chr(i) for i in range(50, 60, 2)]
    # })
    # doc = generate_table(content_s=content_s, orientation="v", chunk_size=4, b_allow_misaligned_values=True,
    #                      f_gen_order_of_values=lambda x: (-int(x["y/n"] is False), -(x["a"] % 3)))
    # print(doc)
    # import os
    #
    # with open(os.path.join(
    #         "/home/SENSETIME/xukaiming/Desktop/my_repos/python_projects/kevin_toolbox/kevin_toolbox/data_flow/file/markdown/test/test_data/for_generate_table",
    #         f"data_5.md"), "w") as f:
    #     f.write(doc)



import os
import torch
from kevin_toolbox.nested_dict_list.serializer.backends import Backend_Base
from kevin_toolbox.nested_dict_list.serializer.variable import SERIALIZER_BACKEND


@SERIALIZER_BACKEND.register()
class Torch_Tensor(Backend_Base):
    name = ":torch:tensor"

    def write(self, name, var, **kwargs):
        assert self.writable(var=var)

        torch.save(var, os.path.join(self.paras["folder"], f'{name}.pt'))
        details = dict(shape=list(var.shape), device=var.device.type, dtype=f'{var.dtype}')
        return dict(backend=Torch_Tensor.name, name=name, details=details)

    def read(self, name, **kwargs):
        assert self.readable(name=name)

        return torch.load(os.path.join(self.paras["folder"], f'{name}.pt'))

    def writable(self, var, **kwargs):
        """
            是否可以写
        """
        return torch.is_tensor(var)

    def readable(self, name, **kwargs):
        """
            是否可以写
        """
        return os.path.isfile(os.path.join(self.paras["folder"], f'{name}.pt'))


if __name__ == '__main__':
    backend = Torch_Tensor(folder=os.path.join(os.path.dirname(__file__), "temp"))

    a = torch.randn(100, device=torch.device("cuda"))
    print(backend.write(name=":a:b", var=a))

    b = backend.read(name=":a:b")
    print(b)

    print(SERIALIZER_BACKEND.database)



import re
from kevin_toolbox.nested_dict_list import get_nodes


def parse_references(var, flag="v"):
    """
        解释 var 中包含引用的值
            什么是引用？
                对于值，若为字符串类型，且其中含有 "...<flag>{ref_name}..." 的形式，则表示解释该值时需要将 <flag>{ref_name} 这部分
                替换为 var 中 ref_name 对应的值

        参数：
            var:
            flag:               <str> 引用标记头
                                    默认为 "v"
        返回：
            {name :{"expression": <str>, "paras": <dict of paras>}, ...}

        示例：
            比如对于：
                name=":z", value="<v>{:x}+<v>{:y}"
            的节点，将会返回：
                {":z": {"expression":"p_0+p_1" , "paras": {"p_0":":x","p_1":":y"}}, ...}
            利用 "expression" 和 "paras" 中的内容，将可以很方便得使用 eval() 和 get_value() 完成对节点值的计算。
            但是由于节点之间可能存在相互引用，因此一般需要通过 cal_relation_between_references() 来确定计算顺序。
    """
    pattern = f'<{flag}>' + r"\{([^{}]+?)\}"

    node_s = dict()
    for name, value in get_nodes(var=var, level=-1, b_strict=True):
        if not isinstance(value, (str,)):
            continue
        ref_names = set(re.findall(pattern=pattern, string=value))
        if len(ref_names) == 0:
            continue
        # 尝试将表达式中的 ref_name 对应部分替换成 p_x 形式的变量，方便后面使用 eval 来解释
        #       比如对于 "<v>{cfg:x}+<v>{cfg:y}" 将会产生表达式 "p_0+p_1" , {"p_0":":x","p_1":":y"}
        prefix = "p_"
        while prefix in value:
            prefix += "_"
        paras = dict()
        expression = value
        for i, var_name in enumerate(ref_names):
            p = f'{prefix}{i}'
            expression = expression.replace(f'<{flag}>{{{var_name}}}', p)
            paras[p] = var_name
        #
        node_s[name] = dict(expression=expression, paras=paras)

    return node_s



import math


def chunk_generator(inputs, chunk_size, b_drop_last=False, b_display_progress=False, hook_for_progress_rate=None):
    """
        构建返回指定批大小的生成器

        参数：
            inputs:                     <Sequence/iterator/generator>
            chunk_size:                 <int> 批大小
            b_drop_last:                <boolean> 若最后一个 chunk 的大小不满足指定的 chunk_size，则抛弃
                                            默认为 False，不开启
            b_display_progress:         <boolean> 显示进度条
            hook_for_progress_rate:     <dict/None> 用于获取进度的“钩子”
                                            当其不为 None 时，将在每次迭代返回 batch 时，同步修改钩子中 "progress_rate" 对应的值为当前进度，
                                                当 inputs 的大小已知，进度是一个 0~1 的 float 值，
                                                当大小未知，进度是一个大于 0 的整数，表示当前是第几个 batch。
    """
    assert chunk_size > 0
    if b_display_progress:
        from tqdm import tqdm
        inputs = tqdm(inputs)

    if isinstance(hook_for_progress_rate, (dict,)):
        count = 0
        if hasattr(inputs, "__len__"):
            total = len(inputs) / chunk_size
            total = math.floor(total) if b_drop_last else math.ceil(total)

            def update_progress_rate():
                nonlocal total, count, hook_for_progress_rate
                count += 1
                hook_for_progress_rate["progress_rate"] = count / total
        else:
            def update_progress_rate():
                nonlocal count, hook_for_progress_rate
                count += 1
                hook_for_progress_rate["progress_rate"] = count
    else:
        def update_progress_rate():
            pass

    #
    chunk = []
    for it in inputs:
        chunk.append(it)
        if len(chunk) == chunk_size + 1:
            update_progress_rate()
            yield chunk[:chunk_size]
            chunk = chunk[-1:]

    if b_drop_last and len(chunk) < chunk_size:
        return
    else:
        update_progress_rate()
        yield chunk


if __name__ == '__main__':
    hook_for_progress_rate_ = dict()
    for i in chunk_generator(inputs=range(2), chunk_size=3, hook_for_progress_rate=hook_for_progress_rate,
                             b_display_progress=True, b_drop_last=False):
        print(hook_for_progress_rate_, i)



import os
import time
from kevin_toolbox.patches import for_os
from kevin_toolbox.data_flow.file import json_
import kevin_toolbox.nested_dict_list as ndl


def read(input_path, **kwargs):
    """
        读取 input_path 中保存的嵌套字典列表

        参数：
            input_path:             <path> 文件夹或者 .tar 文件，具体结构参考 write()
    """
    from kevin_toolbox.nested_dict_list.serializer.variable import SERIALIZER_BACKEND

    assert os.path.exists(input_path)

    # 解压
    temp_dir = None
    if os.path.isfile(input_path) and input_path.endswith(".tar"):
        while True:
            temp_dir = os.path.join(os.path.dirname(input_path), f'temp{time.time()}')
            if not os.path.isdir(temp_dir):
                os.makedirs(temp_dir)
                break
        for_os.unpack(source=input_path, target=temp_dir)
        input_path = os.path.join(temp_dir, os.listdir(temp_dir)[0])

    # 读取 var
    var = json_.read(file_path=os.path.join(input_path, "var.json"), b_use_suggested_converter=True)

    # 读取被处理的节点
    processed_nodes = []
    if os.path.isfile(os.path.join(input_path, "record.json")):
        for name, value in ndl.get_nodes(
                var=json_.read(file_path=os.path.join(input_path, "record.json"),
                               b_use_suggested_converter=True)["processed"], level=-1, b_strict=True):
            if value:
                processed_nodes.append(name)
    else:
        def converter(idx, value):
            processed_nodes.append(idx)
            return value

        ndl.traverse(
            var=var,
            match_cond=lambda _, __, value: isinstance(value, (dict,)) and "backend" in value and "name" in value,
            action_mode="replace", converter=converter, b_use_name_as_idx=True, traversal_mode="bfs",
            b_traverse_matched_element=False)

    # 恢复被处理的节点
    for name in processed_nodes:
        value = ndl.get_value(var=var, name=name)
        if isinstance(value, (dict,)) and "backend" in value and "name" in value:
            bk = SERIALIZER_BACKEND.get(name=value.pop("backend"))(folder=os.path.join(input_path, "nodes"))
            ndl.set_value(var=var, name=name, value=bk.read(**value))

    #
    if temp_dir is not None:
        for_os.remove(path=temp_dir, ignore_errors=True)

    return var


if __name__ == '__main__':
    res = read(
        "/home/SENSETIME/xukaiming/Desktop/my_repos/python_projects/kevin_toolbox/kevin_toolbox/nested_dict_list/serializer/temp3.tar")
    print(res)



import os
import optuna
from kevin_toolbox.data_flow.file import json_
from kevin_toolbox.patches.for_optuna import build_sampler, build_storage


def build_study(output_dir=None, feasible_domain_path=None, **kwargs):
    """
        参数：
            storage:            <dict/str> 数据库链接
                                    有两种指定方式：
                                        1. <str> 直接使用给定的 URI 链接
                                        2. <dict> 使用 build_storage() 根据参数键值对构建数据库
                                    建议使用第2种方式，更具通用性，并且在 mysql 模式下支持新建数据库。
            sampler:            <dict> 采样优化算法
                                    将使用 build_sampler() 进行构建

            output_dir:         <path> 用于补充 build_storage() 中的 output_dir 参数
            feasible_domain_path:   <path> 用于补充 build_sampler() 中的 feasible_domain 参数
            （其他参数参见 optuna.create_study() 函数介绍）
    """
    # 处理参数
    #   storage
    if "storage" in kwargs:
        if isinstance(kwargs["storage"], (str,)) and kwargs["storage"].startswith("mysql:"):
            temp_ls = list(kwargs["storage"].split(":", 1))
            temp[0] = "mysql+pymysql"
            kwargs["storage"] = ":".join(temp_ls)
        elif isinstance(kwargs["storage"], (dict,)):
            kwargs["storage"] = build_storage(output_dir=output_dir, **kwargs["storage"]).url
    #   sampler
    if "sampler" in kwargs and isinstance(kwargs["sampler"], (dict,)):
        if feasible_domain_path is not None:
            kwargs["sampler"]["feasible_domain"] = json_.read(file_path=feasible_domain_path,
                                                              b_use_suggested_converter=True)
        kwargs["sampler"] = build_sampler(**kwargs["sampler"])
    user_attr_s = kwargs.pop("user_attr_s", dict())

    # 创建实验
    study = optuna.create_study(**kwargs)
    for k, v in user_attr_s.items():
        study.set_user_attr(k, v)
    if "storage" not in user_attr_s and "storage" in kwargs:
        study.set_user_attr("storage", kwargs["storage"])

    return study



import optuna
import kevin_toolbox.nested_dict_list as ndl
from kevin_toolbox.patches.for_optuna.serialize import for_trial


def __converter(_, v):
    # trials
    if "trial_id" in v:
        try:
            return for_trial.load(var=v)
        except:
            pass
    # directions/direction
    if v.get("name", None) == "optuna.study.StudyDirection":
        v = eval(v["name"])(v["value"])
    return v


def load(var, study: optuna.study.Study = None):
    """
        加载已序列化的 study
            目前只支持加载以下项目：                额外给定 study 参数时：
                - study_name                    skip
                - directions/direction          check
                - user_attrs                    skip
                - trials                        load
        参数：
            var:                <ndl> 序列化过的 study
            study:              当给定时，将会首先检查该 study 的配置与待加载项目（除trials以外）是否一致，然后再加载 trials
    """
    assert isinstance(var, dict)

    # 反序列化
    var = ndl.traverse(var=var, match_cond=lambda _, __, v: isinstance(v, (dict,)),
                       action_mode="replace", converter=__converter, b_use_name_as_idx=False)

    # 保存到 study 中
    if study is None:
        study = optuna.create_study(
            study_name=var["__dict__"]["study_name"], directions=var["directions"])
        # user_attrs
        for key, value in var["user_attrs"].items():
            study.set_user_attr(key, value)
    else:
        # check directions/direction
        assert study.direction == var["direction"] or study.directions == var["directions"]

    # trials
    for trial in var["trials"]:
        study.add_trial(trial)

    return study



import os
import logging


def build_logger(name, handler_ls, level=logging.DEBUG,
                 formatter="%(asctime)s - %(name)s - %(levelname)s - %(message)s", registry=None):
    """
        构建 logger，并注册到给定的 registry 空间中

        参数：
            name:               <str> 注册名称
            handler_ls:         <list of dict> 日志的输出方式
                                    其中每个 dict 应该包含以下字段
                                        - "target":     输出目标，当为 str 时表示输出文档路径，当为 None 时表示输出到控制台
                                        - "level":      捕获信息的级别
                                        - "formatter":  输出格式
                                    其中 "target" 是必须的，当 "level" 和 "formatter" 置空时，将从全局参数中获取
            level:              <int/str> （全局的）捕获信息的级别
                                    默认为 "DEBUG", 具体参看 logging 中的级别
            formatter:          <str> （全局的）输出格式
            registry:           <Registry> 注册空间
                                    默认不指定。
                                    当指定时，将尝试根据 name 从该空间中获取已有的 logger，或者创建后将新建的 logger 以 name 注册到该空间中。

                    !!注意：避免重复在多个空间中注册，难以管理，本函数所构建的 logger 将会从 logging 中的默认空间（logging.Logger.manager.loggerDict）
                            中删除。
    """
    if registry is not None:
        # 如果在 registry 中已存在，则不再创建，直接返回
        logger = registry.get(name=name, default=None)
        if logger is not None:
            return logger

    if len(handler_ls) == 0:
        return

    # 创建一个logger对象
    logger = logging.getLogger(name)
    logger = logging.Logger.manager.loggerDict.pop(name, logger)  # 从 logging 原生的管理器中删除注册记录，避免重复在多个空间中注册，难以管理
    logger.setLevel(logging.DEBUG)

    # 创建 handler，用于写入日志
    for details in handler_ls:
        if isinstance(details["target"], (str,)):
            # 输出到文件
            os.makedirs(os.path.dirname(details["target"]), exist_ok=True)
            handler = logging.FileHandler(details["target"])
        elif details["target"] is None:
            # 输出到控制台
            handler = logging.StreamHandler()
        else:
            raise ValueError(f'unexpected target {target}')
        handler.setLevel(details.get("level", level))
        handler.setFormatter(logging.Formatter(details.get("formatter", formatter)))
        # 添加到logger中
        logger.addHandler(handler)

    if registry is not None:
        # 将 logger 注册到 registry 中
        registry.add(obj=logger, name=name, b_force=False, b_execute_now=False)

    return logger



from kevin_toolbox.nested_dict_list.name_handler import parse_name
from kevin_toolbox.nested_dict_list import copy_


def cal_relation_between_references(node_s, b_verbose=False):
    """
        计算具有引用的节点之间的关系
            当节点之间的依赖关系满足有向无环图 DAG 时，会输出便于计算节点的顺序 order

        参数：
            node_s:             <dict> parse_references() 返回的结果

        返回：
            node_s, b_is_DAG, order

            node_s:             <dict> 在输入的 node_s 的基础上为每个节点补充 upstream_node 和 downstream_node 字段
                                    其中：
                                        upstream_node 中保存该节点所依赖的上游节点，意味着要等这些上游节点计算完毕后才能计算该节点
                                        downstream_node 中保存对该节点有依赖的下游节点。
            b_is_DAG:           <boolean> 节点关系是否满足有向无环图 DAG
            order:              <list of name> 节点在 DAG 中的顺序
    """

    for v in node_s.values():
        v.update(dict(upstream_node=set(), downstream_node=set()))

    for name_i, details_i in node_s.items():
        for name_j, details_j in node_s.items():
            #
            for ref_name in details_i["paras"].values():
                _, method_ls_r, node_ls_r = parse_name(name=ref_name, b_de_escape_node=False)
                _, method_ls_j, node_ls_j = parse_name(name=name_j, b_de_escape_node=False)
                len_ = min(len(method_ls_r), len(method_ls_j))
                if method_ls_r[:len_] == method_ls_j[:len_] and node_ls_r[:len_] == node_ls_j[:len_]:
                    details_i["upstream_node"].add(name_j)
                    details_j["downstream_node"].add(name_i)
                    break

    # 按照上下游依赖关系进行排序
    upstream_node_names = set()
    downstream_node_s = copy_(var=node_s, b_deepcopy=True)
    order = []
    while True:
        # 将已解释节点放入 order
        order += list(upstream_node_names)
        upstream_node_names.clear()
        # 找出下游节点中已经无依赖的节点，作为新的已解释节点
        for ref_name, details in list(downstream_node_s.items()):
            if len(details["upstream_node"]) == 0:
                upstream_node_names.add(ref_name)
                downstream_node_s.pop(ref_name)
        if len(upstream_node_names) == 0:
            break
        # 清除下游节点中对已解释节点（放置在upstream_node_names中）的依赖
        for details in downstream_node_s.values():
            details["upstream_node"].difference_update(upstream_node_names)
    #
    b_is_DAG = len(downstream_node_s) == 0
    if not b_is_DAG:
        order = None
        if b_verbose:
            print(f'There is a loop in the dependency relationship, and the parsing fails. \n'
                  f'Please check the relationship between the following nodes: {list(downstream_node_s.keys())}')

    return node_s, b_is_DAG, order



from tqdm import tqdm
import optuna


def copy_study(src, dst, b_force=False):
    """
        复制 study
            （本函数改进自 optuna.copy_study，支持更多复制方式）

        参数：
            src:                <dict/optuna.study.Study> 
                                    有两种输入方式：
                                        1. 内存中的 study 实例对象。
                                        2. 从数据库 storage 中读取一个 study 对象。
                                            需要输入包含以下两个键值对的 dict：
                                                - "storage":            <str/optuna.storages.BaseStorage/None> 数据库
                                                                            默认为 None，表示从内存中读取
                                                - "study_name":         <str/None> study 在数据库中的标识符
                                                                            默认为 None
            dst:                <dict> 指定要复制到哪个目标数据库中哪个标识符下
                                    输入方式参考上面的方式 2。
            b_force:            <boolean> 当目标已存在时，是否进行覆盖。
                                    默认为 False，此时若目标已存在则报错。
    """
    # 校验参数
    if not isinstance(src, (optuna.study.Study, dict)):
        raise TypeError("src must be a dict or optuna.study.Study.")
    if not isinstance(dst, (dict,)):
        raise TypeError("dst must be a dict.")

    # 读取 src 对应的 study
    if isinstance(src, (dict,)):
        src_name = src.get("study_name", None)
        src = optuna.load_study(study_name=src_name, storage=src.get("storage", None))
    else:
        src_name = src.study_name

    # 构建 dst 的 study 实例
    dst_name = dst.get("study_name", src_name)
    if b_force:
        try:
            optuna.delete_study(study_name=dst_name, storage=dst.get("storage", None))
        except:
            pass
    dst = optuna.create_study(
        study_name=dst_name, storage=dst.get("storage", None),
        directions=src.directions, load_if_exists=False
    )

    # 复制
    for key, value in src._storage.get_study_system_attrs(src._study_id).items():
        dst._storage.set_study_system_attr(dst._study_id, key, value)
    for key, value in src.user_attrs.items():
        dst.set_user_attr(key, value)
    # trials
    for trial in tqdm(src.get_trials(deepcopy=False)):
        dst.add_trial(trial)  # 之所以不用 add_trials 而是用 add_trial 逐个添加，是为了在使用 sql 数据库时更加鲁棒

    return dst



from kevin_toolbox.computer_science.algorithm.statistician import init_var
import kevin_toolbox.nested_dict_list as ndl


class Accumulator_Base(object):
    """
        累积型统计器的抽象基类

        包含以下变量：
            self.var            <torch.tensor / np.ndarray / int / float> 用于保存累积值
                                    使用 statistician._init_var() 函数进行初始化。
                                    有三种初始化累积值（指定输入数据的格式）方式：
                                        1. （在初始化实例时）显式指定数据的形状和所在设备等。
                                            data_format:        <dict of paras>
                                                    其中需要包含以下参数：
                                                        type_:              <str>
                                                                                "numpy":        np.ndarray
                                                                                "torch":        torch.tensor
                                                        shape:              <list of integers>
                                                        device:             <torch.device>
                                                        dtype:              <torch.dtype>
                                        2. （在初始化实例时）根据输入的数据，来推断出形状、设备等。
                                            like:               <torch.tensor / np.ndarray / int / float>
                                        3. 不在初始化实例时指定 data_format 和 like，此时将等到第一次调用 add()/add_sequence()
                                            时再根据输入来自动推断。
                                            要实现该方式，需要在 add() 中添加:
                                                if self.var is None:
                                                    self.var = self._init_var(like=var)
                                        以上三种方式，默认选用最后一种。
                                        如果三种方式同时被指定，则优先级与对应方式在上面的排名相同。
            self.state          <dict> 用于保存状态的字典
                                    其中包含以下字段：
                                        "total_nums":     <int> 用于统计一共调用了多少次 self.add() 方法去进行累积
            self.paras          <dict> 用于保存构建实例时的各个参数

        包含以下接口：
            add()               (*)添加单个数据
            add_sequence()      (*)添加一系列数据
            get()               (*)获取累积值
            clear()             情况已有数据（self.var）和状态（self.state）
            state_dict()        返回当前实例的状态（返回一个包含 self.var 和 self.state 的字典）
            load_state_dict()   通过接受 state_dict 来更新当前实例的状态

        有可能需要另外实现or覆写的函数：
            上面带有 (*)的接口
            _init_state() 方法
            _init_var() 方法
    """

    def __init__(self, **kwargs):
        """
            至少包含以下参数：
                data_format:            指定数据格式（对应方式1）
                like:                   指定数据格式（对应方式2）
        """
        # 默认参数
        paras = {
            # 指定输入数据的形状、设备
            "data_format": None,
            "like": None,
        }

        # 获取参数
        paras.update(kwargs)

        # 校验参数
        #
        self.paras = paras
        self.var = self._init_var(like=paras["like"], data_format=paras["data_format"])
        self.state = self._init_state()

    def add_sequence(self, var_ls, **kwargs):
        # for var in var_ls:
        #     self.add(var)
        raise NotImplementedError

    def add(self, var, **kwargs):
        """
            添加单个数据

            参数:
                var:                数据
        """
        # if self.var is None:
        #     self.var = self._init_var(like=var)
        # # 累积
        # self.state["total_nums"] += 1
        # # 对 self.var 做处理（需要具体实现）
        raise NotImplementedError

    def get(self, **kwargs):
        """
            获取当前累加的平均值
                当未有累积时，返回 None
        """
        # if len(self) == 0:
        #     return None
        # # 对 self.var 做处理并返回（需要具体实现）
        raise NotImplementedError

    @staticmethod
    def _init_state():
        """
            初始化状态
        """
        return dict(
            total_nums=0,
        )

    @staticmethod
    def _init_var(like=None, data_format=None):
        if like is not None:
            var = init_var.by_like(var=like)
        elif data_format is not None:
            var = init_var.by_data_format(**kwargs)
        else:
            var = None
        return var

    def clear(self):
        self.var = self._init_var(like=self.var)
        self.state = self._init_state()

    def __len__(self):
        return self.state["total_nums"]

    # ---------------------- 用于保存和加载状态 ---------------------- #

    def load_state_dict(self, state_dict):
        """
            加载状态
        """
        self.clear()
        self.state.update(state_dict.get("state", dict()))
        if state_dict.get("var", None) is not None:
            if self.var is None:
                self.var = state_dict["var"]
            else:
                self.var *= 0
                self.var += state_dict["var"]

    def state_dict(self):
        """
            获取状态
        """
        return ndl.copy_(var={"state": self.state, "var": self.var}, b_deepcopy=True, b_keep_internal_references=True)



import os
from kevin_toolbox.data_flow.file import json_
from kevin_toolbox.nested_dict_list.serializer.backends import Backend_Base
from kevin_toolbox.nested_dict_list.serializer.variable import SERIALIZER_BACKEND


@SERIALIZER_BACKEND.register()
class Json_(Backend_Base):
    name = ":json"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.w_cache = None
        self.w_cache_s = dict(able=None, id_=None, content=None)

    def write(self, name, var, **kwargs):
        assert self.writable(var=var)

        with open(os.path.join(self.paras["folder"], f'{name}.json'), "w") as f:
            f.write(self.w_cache_s["content"])

        self.w_cache_s["content"], self.w_cache_s["id_"] = None, None
        return dict(backend=Json_.name, name=name)

    def read(self, name, **kwargs):
        assert self.readable(name=name)

        var = json_.read(file_path=os.path.join(self.paras["folder"], f'{name}.json'), b_use_suggested_converter=True)
        return var

    def writable(self, var, **kwargs):
        """
            是否可以写
        """
        if id(var) != self.w_cache_s["id_"]:
            try:
                self.w_cache_s["content"] = json_.write(content=var, file_path=None, sort_keys=False,
                                                        b_use_suggested_converter=True)
            except:
                self.w_cache_s["content"], self.w_cache_s["id_"] = None, None
                self.w_cache_s["able"] = False
            else:
                self.w_cache_s["id_"] = None
                self.w_cache_s["able"] = True
        return self.w_cache_s["able"]

    def readable(self, name, **kwargs):
        """
            是否可以写
        """
        return os.path.isfile(os.path.join(self.paras["folder"], f'{name}.json'))


if __name__ == '__main__':
    backend = Json_(folder=os.path.join(os.path.dirname(__file__), "temp"))

    var_ = [{123: 123, None: None, "<eval>233": 233, "foo": (2, 3, 4)}, 233]
    # backend.writable(var=var_)
    print(backend.write(name=":inst", var=var_))

    b = backend.read(name=":inst")
    print(b)



from kevin_toolbox.nested_dict_list import traverse, get_value
from kevin_toolbox.nested_dict_list.name_handler import parse_name


def get_nodes(var, level=-1, b_strict=True):
    """
        获取嵌套字典列表 var 中所有叶节点
            以列表 [(name,value), ...] 形式返回，其中名字 name 的解释方式参考 name_handler.parse_name() 介绍

        参数：
            var:                待处理数据
                                    当 var 不是 dict 或者 list 时，返回空列表
            level:              <int> 获取第几层节点
                                    默认为 -1，获取叶节点
            b_strict:           <boolean> 当树结构的某个分枝上的深度不满足 level 的要求时，是否将路径末端的叶节点（当level为正数）或者
                                    路径首部的根节点（当level为负数）添加到输出中。
                                    例如，对于 {'d': {'c': 4}, 'c': 4}，
                                        当 b_strict=True 时，
                                            对于 level=-10 和 level=10 返回的都是 []
                                        当 b_strict=False 时，
                                            对于 level=-10 返回的是 [('', {'d': {'c': 4}, 'c': 4}), ]
                                            对于 level=10 返回的是 [(':c', 4), (':d:c', 4)]
                                    默认为 True，不添加。
    """
    assert isinstance(level, (int,))
    if level == 0:
        return [("", var)]
    res = []

    # 首先找出所有叶节点 level=-1
    def func(_, idx, v):
        nonlocal res
        if not isinstance(v, (list, dict,)):
            res.append((idx, v))
        return False

    traverse(var=var, match_cond=func, action_mode="skip", b_use_name_as_idx=True)

    if level != -1:
        names = set()
        if level < -1:
            for name, _ in res:
                root_node, _, node_ls = parse_name(name=name, b_de_escape_node=False)
                node_ls.insert(0, root_node)
                temp = [len(i) for i in node_ls[level + 1:]]
                temp = len(name) - sum(temp) - len(temp)
                if b_strict and temp < 0:
                    continue
                else:
                    temp = max(temp, 0)
                names.add(name[:temp])
        elif level > 0:
            for name, _ in res:
                root_node, _, node_ls = parse_name(name=name, b_de_escape_node=False)
                node_ls.insert(0, root_node)
                temp = [len(i) for i in node_ls[:level + 1]]
                temp = sum(temp) + level
                if b_strict and temp > len(name):
                    continue
                names.add(name[:temp])
        else:
            raise ValueError
        res.clear()
        for name in names:
            res.append((name, get_value(var=var, name=name)))

    return res


if __name__ == '__main__':
    import numpy as np

    x = [dict(d=3, c=4), np.array([[1, 2, 3]])]
    print(get_nodes(var=x, level=-1))
    print(get_nodes(var=x, level=-2))
    print(get_nodes(var=x, level=-3))
    print(get_nodes(var=x, level=-4))

    print(get_nodes(var=x, level=0))
    print(get_nodes(var=x, level=1))
    print(get_nodes(var=x, level=2))
    print(get_nodes(var=x, level=3))
    print(get_nodes(var=x, level=4))
    print(get_value(var=x, name=""))



from kevin_toolbox.nested_dict_list.serializer.backends import Backend_Base
from kevin_toolbox.nested_dict_list.serializer.variable import SERIALIZER_BACKEND


@SERIALIZER_BACKEND.register(name=":skip:simple")
class Skip_Simple(Backend_Base):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.w_cache_s = dict(able=None, id_=None)

    def write(self, name, var, **kwargs):
        assert self.writable(var=var)
        return var

    def read(self, name, **kwargs):
        raise Exception(f'calling skip.read() is prohibited')

    def writable(self, var, **kwargs):
        """
            是否可以写
        """
        if id(var) != self.w_cache_s["id_"]:
            self.w_cache_s["able"] = False
            if isinstance(var, (int, float, str, type(None))):
                self.w_cache_s["able"] = True
            elif isinstance(var, (tuple,)):
                self.w_cache_s["able"] = all(isinstance(i, (int, float, str, type(None))) for i in var)
            self.w_cache_s["id_"] = id(var)
        return self.w_cache_s["able"]

    def readable(self, name, **kwargs):
        """
            是否可以写
        """
        return False


if __name__ == '__main__':
    import os

    backend = Skip_Simple(folder=os.path.join(os.path.dirname(__file__), "temp"))

    a = None
    print(backend.write(name=":a:b", var=a))

    b = backend.read(name=":a:b")
    print(b)



from kevin_toolbox.nested_dict_list import get_value
from kevin_toolbox.nested_dict_list.name_handler import parse_name, escape_node


def set_value(var, name, value, b_force=False):
    """
        通过解释名字得到取值方式，然后到 var 中将对应部分的值修改为 value。

        参数：
            var:            任意支持索引赋值的变量
            name:           <string> 名字
                                名字 name 的具体介绍参见函数 name_handler.parse_name()
                                假设 var=dict(acc=[0.66,0.78,0.99])，如果你想将 var["acc"][1] 设置为 100，那么可以将 name 写成：
                                    ":acc@1" 或者 "|acc|1" 等。
                                注意，在 name 的开头也可以添加任意非取值方式的字符，本函数将直接忽略它们，比如下面的:
                                    "var:acc@1" 和 "xxxx|acc|1" 也能正常写入。
            value:          待赋给的值
            b_force:        <boolean> 当无法设置时，是否尝试创建或者修改节点
                                默认为 False，此时若无法设置，则报错
                                当设置为 True，可能会对 var 的结构产生不可逆的改变，请谨慎使用。
                                    - 根据取值方式的不同，新创建或者修改的节点的类型可能是 dict 或者 list，
                                        对于 list，其中缺省值填充 None。
                                    - 当需要创建节点时，| 方式将优先创建 dict
                注意：
                    若 b_force 为 True 有可能不会在 var 的基础上进行改变，而是返回一个新的ndl结构，
                    因此建议使用赋值 var = ndl.set_value(var) 来避免可能的错误。
    """
    _, method_ls, node_ls = parse_name(name=name, b_de_escape_node=False)
    if len(node_ls) == 0:
        return value

    raw_key = escape_node(node=node_ls[-1], b_reversed=True, times=1)

    try:
        item = get_value(var=var, name=name[:-1 - len(node_ls[-1])])
        if method_ls[-1] == "@":
            key = eval(raw_key)
        elif method_ls[-1] == "|":
            try:
                _ = item[raw_key]
                key = raw_key
            except:
                key = eval(raw_key)
        else:
            key = raw_key

        if isinstance(item, (list,)) and isinstance(key, (int,)) and len(item) <= key:
            item.extend([None] * (key - len(item) + 1))
        item[key] = value
    except:
        if not b_force:
            raise ValueError(f'The location pointed to by name {name} does not exist in var')
        else:
            if method_ls[-1] in "|:":
                # 对于字符串默认使用 dict 构建
                value = {raw_key: value}
            else:
                # 对于需要eval的情况
                key = eval(raw_key)
                if isinstance(key, (int,)) and key >= 0:
                    # 只有当 key 为非负整数时，才会使用 list 构建
                    value = [None] * key + [value]
                else:
                    # 其他，比如当 key 为元组、浮点数等等时，则使用 dict 构建
                    value = {key: value}
            var = set_value(var=var, name=name[:-1 - len(node_ls[-1])], value=value, b_force=b_force)

    return var



import os
from collections import defaultdict
from enum import Enum


class Ignore_Scope(Enum):
    ROOT = "root"
    DIRS = "dirs"
    FILES = "files"


class Path_Ignorer:
    def __init__(self, ignore_s=None):
        self.ignore_s = defaultdict(list)
        try:
            if isinstance(ignore_s, (list, tuple)):
                for it in ignore_s:
                    for scope in it["scope"]:
                        scope = Ignore_Scope(scope)
                        self.ignore_s[scope].append(it["func"])
            elif isinstance(ignore_s, (dict,)):
                for k, v in ignore_s.items():
                    self.ignore_s[Ignore_Scope(k)].extend(v)
        except:
            raise ValueError(f'invalid ignore_s, got a {ignore_s}')

    def __call__(self, scope, *args, **kwargs):
        for func in self.ignore_s[scope]:
            if func(*args, **kwargs):
                return True
        return False

    def __len__(self):
        return len(self.ignore_s)


def walk(top, topdown=True, onerror=None, followlinks=False, ignore_s=None):
    """
        在 os.walk() 的基础上增加了以下功能
            - 可以通过 ignore_s 参数来排除特定的目录和文件

        参数：
            （对于参数 top,topdown,onerror,followlinks，其用法与 os.walk() 完全相同）
            top:                <path> 要遍历的目录
            topdown:            <boolean> 是否从浅到深遍历
                                    默认为 True，优先遍历完浅层目录
            onerror:            <callable> 当遇到异常时，会调用该函数
            followlinks:        <boolean> 是否遍历软链接目录
                                    默认为 False，不对软链接进行进一步遍历
            ignore_s:           <list/tuple of dict or dict> 排除规则
                                    有两种输入方式：

                                    方式 1 <list/tuple of dict>：
                                        列表中每个字典需要具有以下键值对：
                                            "scope":    <list/tuple of str> 该规则的作用范围
                                                            可选值，及（满足规则时）对应效果：
                                                                "root":         不遍历满足规则的目录
                                                                "dirs":         将返回的三元组中第二个 dirs 部分中满足规则的部分移除
                                                                "files":        将返回的三元组中第三个 files 部分中满足规则的部分移除
                                            "func":     <callable> 排除规则
                                                            当调用该函数的返回值为 True 时，执行排除。
                                                            函数类型为 def(b_is_dir, b_is_symlink, path): ...
                                                            其中：
                                                                b_is_dir        是否是目录
                                                                b_is_symlink    是否是软链接
                                                                path            输入是对于作用范围 "root"，输入的直接就是 root，
                                                                                对于作用范围 "dirs"，输入是 dirs 中的每个元素和 root 组成的绝对路径，
                                                                                对于作用范围 "files"，输入是 files 中的每个元素和 root 组成的绝对路径。
                                            注意：
                                                - 当有多个规则时，只要满足其一即会进行排除。
                                                - 任何规则都不会对 top 目录进行过滤，亦即对 top 的遍历是必然的。不能通过设定某个 top 满足的规则来停止对 top 的遍历。
                                        比如，通过下面的规则就可以实现不对 basename 为 "test" 和 "temp" 的目录进行遍历和输出，
                                        同时只保留非软连接的 .png 和 .jpg 文件：
                                            [
                                                {
                                                    "func": lambda _, __, path: os.path.basename(path) in ["temp", "test"],
                                                    "scope": ["root", "dirs"]
                                                },
                                                {
                                                    "func": lambda _, b_is_symlink, path: b_is_symlink or not path.endswith((".png",".jpg")),
                                                    "scope": ["files", ]
                                                }
                                            ]

                                    方式 2 <dict>:
                                        一个以 scope 为键，func 为值的字典。
                                        延续上面的例子，其另一种等效形式为：
                                            {
                                                "root": [
                                                    lambda _, __, path: os.path.basename(path) in ["temp", "test"],
                                                ],
                                                "dirs": [
                                                    lambda _, __, path: os.path.basename(path) in ["temp", "test"],
                                                ],
                                                "files": [
                                                    lambda _, b_is_symlink, path: b_is_symlink or not path.endswith((".png",".jpg")),
                                                ]
                                            }

                                    方式 3 <Path_Ignorer>:
                                        根据 ignore_s 构建的 Path_Ignorer

    """
    # 根据 ignore_s 构建 Path_Ignorer
    path_ignorer = ignore_s if isinstance(ignore_s, (Path_Ignorer,)) else Path_Ignorer(ignore_s=ignore_s)

    #
    yield from __walk(top, topdown, onerror, followlinks, path_ignorer)


def __walk(top, topdown, onerror, followlinks, path_ignorer):
    #
    top = os.fspath(top)
    dirs, files = [], []
    walk_dirs = []

    try:
        scandir_it = os.scandir(top)
    except OSError as error:
        if onerror is not None:
            onerror(error)
        return

    with scandir_it:
        while True:
            try:
                try:
                    entry = next(scandir_it)
                except StopIteration:
                    break
            except OSError as error:
                if onerror is not None:
                    onerror(error)
                return

            try:
                is_dir = entry.is_dir()
            except OSError:
                is_dir = False

            if is_dir:
                dirs.append(entry.name)
                walk_dirs.append(os.path.join(top, entry.name))
            else:
                files.append(entry.name)

    # 过滤
    if len(path_ignorer) > 0:
        for it_ls, scope in zip([walk_dirs, dirs, files], [Ignore_Scope.ROOT, Ignore_Scope.DIRS, Ignore_Scope.FILES]):
            for i in reversed(range(len(it_ls))):
                path = os.path.join(top, it_ls[i]) if scope != Ignore_Scope.ROOT else it_ls[i]
                if path_ignorer(scope, scope != Ignore_Scope.FILES, os.path.islink(path), path):
                    it_ls.pop(i)

    #
    if topdown:
        yield top, dirs, files
        #
        for new_path in walk_dirs:
            if followlinks or not os.path.islink(new_path):
                yield from __walk(new_path, topdown, onerror, followlinks, path_ignorer)
    else:
        for new_path in walk_dirs:
            if followlinks or not os.path.islink(new_path):
                yield from __walk(new_path, topdown, onerror, followlinks, path_ignorer)
        #
        yield top, dirs, files



from kevin_toolbox.nested_dict_list import get_value, set_value


def set_default(var, name, default, b_force=False, cache_for_verified_names=None):
    """
        当 name 指向的位置在 var 中不存在时，将会把 default 插入到对应的位置。
            （类似于 dict.setdefault() 的行为）

        参数：
            var:            任意支持索引赋值的变量
            name:           <string> 名字（位置）
                                名字 name 的具体介绍参见函数 name_handler.parse_name()
            default:        默认值
            b_force:        <boolean> 当无法将 default 设置到 name 指定的位置时，是否尝试创建或者修改 var 中的节点
                                默认为 False，此时若无法设置，则报错
                                当设置为 True，可能会对 var 的结构产生不可逆的改变，请谨慎使用。
                注意：
                    若 b_force 为 True 有可能不会在 var 的基础上进行改变，而是返回一个新的ndl结构，
                    因此建议使用赋值 var = ndl.set_default(var) 来避免可能的错误。

            cache_for_verified_names:   <set> 用于缓存已检验的 name
                                默认为 None，不使用缓存。
                                当设为某个集合时，将开启缓存。此时将首先判断 name 是否在缓存中，若在，则视为之前已经对该 name 成功进行过 set_default()
                                    操作，没有必要再重复执行，因此直接跳过后续流程；若不在，则执行 set_default() 操作，并在成功执行之后将该 name
                                    补充到缓存中。
                                合理利用该缓存机制将可以避免对同一个 name 反复进行 set_default() 操作，从而提高效率。
    """
    if cache_for_verified_names is not None:
        assert isinstance(cache_for_verified_names, (set,))
        if name in cache_for_verified_names:
            return var

    temp = list()
    if id(temp) == id(get_value(var=var, name=name, default=temp)):
        # 获取失败，说明 name 指向的位置不存在，则尝试创建
        var = set_value(var=var, name=name, value=default, b_force=b_force)

    if cache_for_verified_names is not None:
        cache_for_verified_names.add(name)
    return var



import torch
import numpy as np


def init_by_data_format(type_, shape, **kwargs):
    """
        构建一个与输入 var 具有相同类型、形状、设备的 0 数组

        参数：
            type_:              <str>
                                    "numpy":        np.ndarray
                                    "torch":        torch.tensor
                                    "number":       float
            shape:              <list of integers>
            device:             <torch.device>
            dtype:              <torch.dtype / np.dtype>
    """
    if type_ == "torch":
        res = torch.zeros(size=shape, **kwargs)
    elif type_ == "numpy":
        res = np.zeros(shape=shape, **kwargs)
    else:
        res = 0.0
    return res



from kevin_toolbox.nested_dict_list.name_handler import parse_name


def get_value(var, name, b_pop=False, **kwargs):
    """
        通过解释名字得到取值方式，然后到 var 中获取对应部分的值。

        参数：
            var:            任意支持索引取值的变量
            name:           <str> 名字
                                名字 name 的具体介绍参见函数 name_handler.parse_name()
                                假设 var=dict(acc=[0.66,0.78,0.99])，如果你想读取 var["acc"][1] => 0.78，那么可以将 name 写成：
                                    ":acc@1" 或者 "|acc|1" 等。
                                注意，在 name 的开头也可以添加任意非解释方式的字符，本函数将直接忽略它们，比如下面的:
                                    "var:acc@1" 和 "xxxx|acc|1" 也能正常读取。
            b_pop:          <boolean> 是否将值从 var 中移除
                                默认为 False
            default:        默认值
                                - 不设置（默认）。当取值失败时将报错。
                                - 设置为任意值。取值失败时将返回该值。
    """
    _, method_ls, node_ls = parse_name(name=name, b_de_escape_node=True)

    try:
        pre, cur = None, var
        node = None
        for method, node in zip(method_ls, node_ls):
            pre, cur = cur, None
            if method == "@":
                node = eval(node)
            elif method == "|":
                try:
                    _ = pre[node]
                except:
                    node = eval(node)
            cur = pre[node]

        if b_pop and len(node_ls) > 0:
            assert isinstance(pre, (dict, list,)), \
                f'pop is only supported when the parent node type is list or dict, but got a {type(pre)}'
            pre.pop(node)
    except:
        if "default" in kwargs:
            cur = kwargs["default"]
        else:
            raise IndexError(f'invalid name {name}')

    return cur



import optuna
import kevin_toolbox.nested_dict_list as ndl


def sample_from_feasible_domain(var, trial: optuna.trial.BaseTrial, f_p_name_builder=None, b_use_name_as_idx=True):
    """
        使用试验 trial 基于输入中的定义域 feasible_domain 部分进行参数采样和替换。
            遍历输入中的所有元素，找出符合 <feasible_domain> 格式要求的记录了参数定义域的元素，
            然后使用输入的试验实例 trial 结合参数定义域采样出对应的参数，最后用采样出来的参数替换掉原来的定义域。

        参数：
            var:                <list/dict> 当其中的元素满足 <feasible_domain> 格式要求，将进行采样与替换。
                                    <feasible_domain> 格式要求：
                                        1. 是一个 dictionary
                                        2. 包含 "p_type"  字段
                                            "p_type" 表示定义域类型，常见值包括："float" "int" "categorical" 等
                                        3. 根据不同的定义域类型，给出定义域的参数 
                                            比如，p_type="categorical" 时应该包含可选值列表 "choices"
                                            p_type="float"或者"int" 时应该包含最大、最小、间隔值、坐标轴类型 "high" "low" "step" "log" 等
                                        更多参见 https://optuna.readthedocs.io/zh_CN/latest/reference/generated/optuna.trial.Trial.html#optuna.trial.Trial
                                        中的 suggest_xxx() 函数。
            trial:              <optuna.trial.BaseTrial> 试验
            f_p_name_builder:   <callable> 对于采样出的参数，使用该函数构建其在试验 trial 中注册的名称。
                                    函数类型为 def(idx, p_type): ...
                                    其中：
                                        idx             被采样的节点在 var 中的位置
                                                            当 b_use_name_as_idx=False 时，
                                                                对于列表是 index，对于字典是 key
                                                            当为 True 时，传入的是元素在整体结构中的 name 位置，name的格式和含义参考
                                                                name_handler.parse_name() 中的介绍
                                        p_type          被采样的节点中的 "p_type" 字段
                                    默认为：
                                        lambda idx, p_type: idx
            b_use_name_as_idx:  <boolean> 决定传入 f_p_name_builder 中的参数 idx 的形式。

        返回：
            var, node_vs_paras_s
                var 是采样后的结果
                node_vs_paras_s 是一个<dict>，以被采样的节点在 var 中位置的名称作为键，以对应节点在 trial 中注册的参数名为值。

        实例：
            对于输入 var={
                            "thr":[
                                "interval_thr": {
                                    "p_type": "categorical",
                                    "choices": [
                                      1000,
                                      2000,
                                      5000,
                                      10000
                                    ]
                                },
                                "iou_thr": {
                                    "p_type": "float",
                                    "low": 0,
                                    "high": 1.0000001,
                                    "step": 0.05
                                },
                                "connection": {
                                    "p_type": "categorical",
                                    "choices": {
                                        "skip": lambda x:x,
                                        "and":  lambda x,y : x and y,
                                        "or":  lambda x,y : x or y,
                                    }
                                },
                            ]
                        }
            可能返回的采样结果是 res={"thr":[{"interval_thr":1000}, {"iou_thr":0.6}, {"connection":lambda x:x} ], }。
            当 f_p_name_builder=lambda idx, p_type: f'my{idx}' 时，
            这些参数在 trial 中注册的名称分别是 "my:thr@0:interval_thr"，"my:thr@1:iou_thr" 和 "my:thr@1:connection"。
            特别地，
                - 对于字典形式的 choices，其中保存在 trial 中的取值是其键 key 而非 value。
                - 对于list形式，但含有非支持类型的  choices，其中保存在 trial 中的取值是元素的 index。
            这些名称的含义详见 get_value()。
    """
    if f_p_name_builder is None:
        f_p_name_builder = lambda idx, p_type: idx
    if not b_use_name_as_idx:
        p_name_builder = lambda idx, p_type: f_p_name_builder(ndl.name_handler.parse_name(idx)[-1][-1], p_type)
    else:
        p_name_builder = f_p_name_builder

    node_vs_paras_s = dict()
    p_name_set = set()

    def func(idx, v):
        nonlocal node_vs_paras_s, p_name_builder

        p_type = v.pop("p_type")
        p_name = v.pop("p_name") if "p_name" in v else f'{p_name_builder(idx, p_type)}'
        assert p_name not in p_name_set, \
            f"p_name={p_name} is duplicated!"
        p_name_set.add(p_name)
        kwargs = v
        #
        choice_values = None
        if p_type == "categorical":
            # optuna 目前的类别元素仅支持 None, bool, int, float 和 str 类型
            #   对于其他类型的元素，比如 list 和 dict，需要替换成对应 index_ls 或者 key_ls
            #   然后再根据建议的 index 或者 key 到 list 和 dict 中取值
            assert "choices" in kwargs
            if isinstance(kwargs["choices"], (list, tuple,)) and \
                    not all([isinstance(i, (bool, int, float, str,)) or i is None for i in kwargs["choices"]]):
                choice_values = kwargs["choices"]
                kwargs["choices"] = list(range(len(kwargs["choices"])))
            elif isinstance(kwargs["choices"], (dict,)):
                choice_values = kwargs["choices"]
                kwargs["choices"] = list(kwargs["choices"].keys())

        v = eval(f'trial.suggest_{p_type}(name=name, **kwargs)', {"trial": trial, "name": p_name, "kwargs": kwargs})
        if choice_values is not None:
            v = choice_values[v]

        #
        node_vs_paras_s[idx] = p_name
        return v

    var = ndl.traverse(var=var, match_cond=lambda _, __, v: isinstance(v, (dict,)) and "p_type" in v,
                       action_mode="replace", converter=func, b_use_name_as_idx=True)

    return var, node_vs_paras_s



import kevin_toolbox.nested_dict_list as ndl
from kevin_toolbox.computer_science.data_structure import Executor
from kevin_toolbox.computer_science.algorithm.for_dict import deep_update
from kevin_toolbox.data_flow.file import markdown


class Random_Node:
    def __init__(self, name, predecessor, successor, **kwargs):
        assert callable(pipeline)

        self.name = name
        self.pipeline = None
        # 对于 relation_s，其中 predecessor 字段下保存有：
        #   {<predecessor_node>: [(<output_name>, <input_name>), ...], ...}
        #   其中：
        #       - <predecessor_node>： 前继节点名
        #       - <output_name>： 从该前继节点输出的结果中取出对应部分
        #       - <input_name>： 将取出的部分填入到本节点的 self.paras 中对应位置上
        # 对于 successor字段，其中保存后继节点的名字
        self.relation_s = {
            "predecessor": dict(),
            "successor": set(),
        }
        self.relation_s = deep_update(stem=self.relation_s,
                                      patch={k: kwargs[k] for k in self.relation_s.keys() if k in kwargs})
        self.paras = {
            "args": list(),
            "f_args": list(),
            "kwargs": dict(),
            "f_kwargs": dict(),
        }
        self.paras = deep_update(stem=self.paras,
                                 patch={k: kwargs[k] for k in self.paras.keys() if k in kwargs})
        self.parse()

    def parse(self):
        def func(idx, value):
            nonlocal self
            predecessor, output_name = value.split("{", 1)[-1].split("}", 1)
            if predecessor not in self.relation_s["predecessor"]:
                self.relation_s["predecessor"][predecessor] = []
            self.relation_s["predecessor"][predecessor].append((output_name, idx))
            return None

        ndl.traverse(
            var=self.paras, match_cond=lambda _, __, value: isinstance(value, (str,)) and value.startswith("<exe>"),
            action_mode="replace", converter=func, b_use_name_as_idx=True
        )

    def run(self, predecessor_output_s):
        for k, v in predecessor_output_s.items():
            for output_name, input_name in self.relation_s["predecessor"][k]:
                ndl.set_value(var=self.paras, name=input_name, value=ndl.get_value(var=v, name=output_name),
                              b_force=True)
        if len(self.paras["f_args"]) == len(self.paras["f_kwargs"]) == 0:
            res = self.pipeline(*self.paras["args"], **self.paras["kwargs"])
        else:
            res = Executor(func=self.pipeline, **self.paras).run()
        return res

    def __repr__(self):
        doc = f'<Simple_Node> with name: {self.name}\n' \
              f'\tpipeline: {self.pipeline}\n' \
              f'\tparas: \n'
        doc += markdown.generate_list(var=self.paras, indent=8)
        doc += f'\trelation_s: \n'
        doc += markdown.generate_list(var=self.relation_s, indent=8)
        return doc


if __name__ == '__main__':
    node = Simple_Node(name="@0", pipeline=lambda x, y: x + y, args=["<exe>{@1}:x"], f_kwargs={"y": "<exe>{@2}"})
    print(node)



from .sample_from_feasible_domain import sample_from_feasible_domain
from .build_sampler import build_sampler
from .build_storage import build_storage
from .build_study import build_study
from .copy_study import copy_study
#
from . import serialize



import os
import optuna
from enum import Enum
from kevin_toolbox.patches.for_os import remove


class Storage_Type(Enum):
    SQLITE = "sqlite"
    MYSQL = "mysql"


def build_storage(mode, b_clear_before_create=False, **kwargs):
    """
        构建数据库实例

        参数：
            mode:                       <str> 数据库的种类
                目前支持以下几种模式。当 mode 设定为以下值时：

                - "sqlite"（python自带库，不需要额外环境配置，但不支持并行化）
                    还需要以下参数：
                        output_dir:     <path> 数据库文件保存目录
                        db_name:        <str> 数据库文件名称

                - "mysql"（需要按照 https://zhuanlan.zhihu.com/p/670547423 中进行配置，支持并行化）
                    还需要以下参数：
                        host:           <str> 主机名
                                            默认为 'localhost'
                        user:           <str> 用户名
                                            默认为 'root'
                        password:       <str> 登录密码
                                            默认缺省不指定
                        db_name:        <str> 数据库名
                        port:           <int> 端口
                                            默认为 3306

            其他共有的参数为：
                b_clear_before_create:  <boolean> 在创建前是否尝试清除已有数据库
                                            默认为 False。
                                            注意！！当进行并行化时，建议最多在主进程中设置一次 True，以免多个进程反复删除数据库引发错误。
    """
    mode = Storage_Type(mode)

    # 构建URI
    if mode is Storage_Type.SQLITE:
        assert "output_dir" in kwargs, \
            f'for mode "sqlite", please specify output_dir'
        file_path = os.path.join(kwargs["output_dir"], kwargs.get("db_name", "study.db"))
        url = f'sqlite:///{file_path}'
        #
        os.makedirs(kwargs["output_dir"], exist_ok=True)
        if b_clear_before_create:
            remove(path=file_path, ignore_errors=True)
    else:  # mode is Storage_Type.MYSQL
        assert "db_name" in kwargs, \
            f'for mode "mysql", please specify db_name'
        kwargs.setdefault("host", "localhost")
        kwargs.setdefault("user", "root")
        kwargs.setdefault("password", "")
        kwargs.setdefault("port", 3306)
        url = f'mysql+pymysql://{kwargs["user"]}:{kwargs["password"]}@' \
              f'{kwargs["host"]}/{kwargs["db_name"]}'
        #
        try:
            import pymysql
        except ImportError:
            raise ImportError('for mode "mysql", please install pymysql')
        conn = pymysql.connect(host=kwargs["host"], user=kwargs["user"], password=kwargs["password"],
                               port=kwargs["port"])
        cursor = conn.cursor()
        if b_clear_before_create:
            cursor.execute(f'DROP DATABASE IF EXISTS {kwargs["db_name"]}')
        cursor.execute(f'CREATE DATABASE IF NOT EXISTS {kwargs["db_name"]}')
        cursor.close()
        conn.close()

    return optuna.storages.RDBStorage(url)



def escape_tuple(x):
    """
        将 tuple 进行转义
            转义：     x ==> f"<eval>{x}"
            反转义：   f"<eval>{x}" ==> x

        为什么要进行转义？
            由于 json 中会将 tuple 作为 list 进行保存，因此在保存过程中会丢失相应信息。
    """
    if isinstance(x, tuple) or (isinstance(x, (str,)) and x.startswith("<eval>")):
        return f'<eval>{x}'
    else:
        return x


if __name__ == '__main__':
    print(escape_tuple((1, 2, "\'1\'")))
    # <eval>(1, 2, "'1'")
    print(escape_tuple("<eval>233"))
    # <eval><eval>233



def generate_list(var, indent=0):
    doc = ""
    if isinstance(var, (dict,)):
        for key, value in var.items():
            if isinstance(value, (dict, list,)):
                doc += f'{" " * indent}- {key}\n'
                doc += generate_list(value, indent + 2)
            else:
                doc += f'{" " * indent}- {key}: {value}\n'
    elif isinstance(var, (list,)):
        for value in var:
            if isinstance(value, (dict, list,)):
                doc += f'{" " * indent}- \n'
                doc += generate_list(value, indent + 2)
            else:
                doc += f'{" " * indent}- {value}\n'
    return doc


if __name__ == '__main__':
    # 嵌套字典数据
    nested_dict = {
        'Category 1': {
            'Item 1': [1, 2, 3],
            'Item 2': 20
        },
        'Category 2': {
            'Item 3': 30,
            'Item 4': 40
        }
    }

    # 生成Markdown列表
    markdown_list = generate_list(nested_dict)

    print(markdown_list)



import os
from kevin_toolbox.nested_dict_list.serializer.backends import Backend_Base
from kevin_toolbox.nested_dict_list.serializer.variable import SERIALIZER_BACKEND
from kevin_toolbox.nested_dict_list.serializer import write, read


@SERIALIZER_BACKEND.register()
class NDL(Backend_Base):
    name = ":ndl"

    def write(self, name, var, **kwargs):
        assert self.writable(var=var)

        write(var=var, output_dir=os.path.join(self.paras["folder"], f'{name}'),
              settings=None, traversal_mode="bfs", b_pack_into_tar=False, **kwargs)
        return dict(backend=NDL.name, name=name)

    def read(self, name, **kwargs):
        assert self.readable(name=name)

        input_path = os.path.join(self.paras["folder"], f'{name}.tar') if os.path.isfile(
            os.path.join(self.paras["folder"], f'{name}.tar')) else os.path.join(self.paras["folder"], f'{name}')

        return read(input_path=input_path, **kwargs)

    def writable(self, var, **kwargs):
        """
            是否可以写
        """
        return True

    def readable(self, name, **kwargs):
        """
            是否可以写
        """
        res = os.path.isfile(os.path.join(self.paras["folder"], f'{name}.tar')) or \
              os.path.isdir(os.path.join(self.paras["folder"], f'{name}'))
        return res


if __name__ == '__main__':
    import torch

    backend = NDL(folder=os.path.join(os.path.dirname(__file__), "temp"))

    a = {"a": 1, "b": [torch.randn(100, device=torch.device("cuda"))]}
    print(backend.write(name=":a:b", var=a))

    b = backend.read(name=":a:b")
    print(b)

    print(SERIALIZER_BACKEND.database)



import kevin_toolbox.nested_dict_list as ndl
from kevin_toolbox.nested_dict_list import value_parser as ndl_vp


def parse_and_eval_references(var, flag="v", converter_for_ref=None, converter_for_res=None):
    """
        解释并替换 var 中包含引用的值
            什么是引用？
                对于值，若为字符串类型，且其中含有 "...<flag>{ref_name}..." 的形式，则表示解释该值时需要将 <flag>{ref_name} 这部分
                替换为 var 中 ref_name 对应的值
            本函数系对 parse_references()，eval_references()以及 cal_relation_between_references.py()的集成，
                具体实现细节请参考这些函数

        参数：
            var:
            flag:                   <str> 引用标记头
                                        默认为 "v"

            在计算包含引用的节点的结果时，将对其执行以下方法：
            converter_for_ref:      <callable> 对被引用节点施加何种处理
                                        形如 def(idx, v): ... 的函数，其中 idx 是被引用节点的名字，v是其值，
                                        返回的结果将替换掉被引用节点中原来的值。
                                        注意：
                                            - 处理后得到的结果将替换掉原引用节点的值。（重要所以说两次）
                                            - 当同一节点被多次引用时，仅会被处理、替换一次。
            converter_for_res:      <callable> 对计算结果施加何种处理
                                        形如 def(idx, v): ... 的函数，其中 idx 是节点的名字，v是计算结果

        返回：
            var, name_ls
                var 是解释后的结果，
                name_ls 是被解释的节点名称，按照解释顺序排列
    """
    node_s = ndl_vp.parse_references(var=var, flag=flag)
    node_s, b_is_DAG, order = ndl_vp.cal_relation_between_references(node_s=node_s, b_verbose=True)
    assert b_is_DAG, \
        f'There is a circular <{flag}> reference in config'
    var = ndl_vp.eval_references(var=var, node_s=node_s, order=order,
                                 converter_for_ref=converter_for_ref, converter_for_res=converter_for_res)
    return var, order



from datetime import datetime
import optuna
import kevin_toolbox.nested_dict_list as ndl


def __converter(_, v):
    if isinstance(v, optuna.distributions.BaseDistribution):
        return dict(name=f'optuna.distributions.{v.__class__.__name__}', paras=v.__dict__)
    elif isinstance(v, optuna.trial.TrialState):
        return dict(name=f'optuna.trial.TrialState', value=v.value, description=v.name)
    elif isinstance(v, datetime):
        return dict(name="datetime", timestamp=v.timestamp(), description=f'{v}')
    return v


def dump(trial):
    assert isinstance(trial, optuna.trial.BaseTrial)

    res_s = dict()
    for k, v in ndl.copy_(var=trial.__dict__, b_deepcopy=True).items():
        if k.startswith("_"):
            k = k[1:]
        res_s[k] = v

    res_s = ndl.traverse(var=res_s, match_cond=lambda _, __, v: not isinstance(v, (list, dict,)),
                         action_mode="replace", converter=__converter, b_use_name_as_idx=False)

    return res_s



def unescape_non_str_dict_key(x):
    """
        将字典中的所有非字符串的 key 进行反转义
            转义：     key ==> f"<eval>{key}"
            反转义：   f"<eval>{key}" ==> key

        为什么要进行转义？
            由于 json 中要求字典的键必须使用字符串进行保存，因此在保存过程中会丢失相应信息。
    """
    if isinstance(x, dict):
        res = dict()
        for k, v in x.items():
            if k.startswith("<eval>"):
                k = k[6:]
                if not k.startswith("<eval>"):
                    k = eval(k)
            res[k] = v
        return res
    else:
        return x


if __name__ == '__main__':
    print(unescape_non_str_dict_key({'<eval>123': 123, '<eval><eval>-1.000': -1.0, '<eval>None': None}))



import os
import json
from kevin_toolbox.data_flow.file.json_.converter import integrate, unescape_tuple, unescape_non_str_dict_key
from kevin_toolbox.nested_dict_list import traverse


def read_json(file_path, converters=None, b_use_suggested_converter=False):
    """
        读取 json file

        参数：
            file_path
            converters:                 <list of converters> 对读取内容中每个节点的处理方式
                                            转换器 converter 应该是一个形如 def(x): ... ; return x 的函数，具体可以参考
                                            json_.converter 中已实现的转换器
            b_use_suggested_converter:  <boolean> 是否使用建议的转换器
                                            建议使用 unescape/escape_non_str_dict_key 和 unescape/escape_tuple 这两对转换器，
                                            可以避免因 json 的读取/写入而丢失部分信息。
                                            默认为 False。
                    注意：当 converters 非 None，此参数失效，以 converters 中的具体设置为准
    """
    assert os.path.isfile(file_path), f'file {file_path} not found'
    if converters is None and b_use_suggested_converter:
        converters = [unescape_tuple, unescape_non_str_dict_key]

    with open(file_path, 'r') as f:
        content = json.load(f)

    if converters is not None:
        converter = integrate(converters)
        content = traverse(var=[content],
                           match_cond=lambda _, __, ___: True, action_mode="replace",
                           converter=lambda _, x: converter(x),
                           b_traverse_matched_element=True)[0]

    return content



from datetime import datetime
import optuna
import kevin_toolbox.nested_dict_list as ndl


def __converter(_, v):
    if v["name"].startswith("optuna.distributions."):
        v = eval(v["name"])(**v["paras"])
    elif v["name"] == "optuna.trial.TrialState":
        v = eval(v["name"])(v["value"])
    elif v["name"] == "datetime":
        v = datetime.fromtimestamp(v["timestamp"])

    return v


def load(var):
    assert isinstance(var, dict)

    var = ndl.traverse(var=var, match_cond=lambda _, __, v: isinstance(v, (dict,)) and "name" in v,
                       action_mode="replace", converter=__converter, b_use_name_as_idx=False)
    var.setdefault("value", None)
    var.setdefault("values", None)

    trial = optuna.trial.FrozenTrial(**var)
    trial._validate()
    return trial



import os

BACKEND = ["tarfile", "os"]


def unpack(source, target=None):
    assert os.path.exists(source)
    target = os.path.dirname(source) if target is None else target
    os.makedirs(target, exist_ok=True)

    try:
        assert "tarfile" in BACKEND
        import tarfile
        # 创建一个 TarFile 对象并打开 tar 文件
        with tarfile.open(source, "r") as tar:
            # 遍历 tar 文件中的每个文件或目录
            for member in tar.getmembers():
                # 如果成员是一个文件，则提取到目标目录中
                if member.isfile():
                    tar.extract(member, target)
                # 如果成员是一个目录，则在目标目录中创建相应的目录结构
                elif member.isdir():
                    os.makedirs(os.path.join(target, member.name), exist_ok=True)
    except:
        assert "os" in BACKEND
        os.system(f'cd {target}; tar -xvf {source}')

    assert os.path.exists(target), \
        f'failed to unpack tar file {source}'



from .cal_relation_between_references import cal_relation_between_references
from .eval_references import eval_references
from .parse_references import parse_references
from .parse_and_eval_references import parse_and_eval_references



from .get_hash import get_hash
from .traverse import traverse
from .get_value import get_value
from .set_value import set_value
from .set_default import set_default
from .count_leaf_node_nums import count_leaf_node_nums
from .get_nodes import get_nodes
from .copy_ import copy_
#
from . import serializer, value_parser, name_handler
